#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
from abc import ABCMeta

python_version = sys.version_info

if python_version < (3, 3):
    from collections import Iterable
else:
    from collections.abc import Iterable

from asq.queryables import Queryable

from .monkey_patches import extend_type


def get_method_proxy(method):
    def proxy(self, *args, **kwargs):
        return method(Queryable(self), *args, **kwargs)

    proxy.__name__ = method.__name__
    proxy.__doc__ = method.__doc__
    return proxy


def get_asq_methods():
    return dict((name, get_method_proxy(attr))
                for name, attr in vars(Queryable).items()
                if not name.startswith('_') and callable(attr))


asq_methods = get_asq_methods()


def patch_type(klass):
    """
    Extends the class with ``asq.queryables.Queryable`` methods.

    Appends all public methods of ``asq.queryables.Queryable`` to the class
    in such a way that a method call:

        >>> instance.select(lambda x: x * 2)

    Will be equal to:

        >>> Queryable(instance).select(lambda x: x * 2)

    If the class already contains an attribute with a coinciding name
    (e.g. str.join and list.count), this attribute won't be replaced.

    Args:
        klass: A class (but not a metaclass) implements
               ``collection.abc.Iterable`` interface. Can be built-in.

    Raises:
        TypeError: If the argument isn't a class that satisfy
            the requirements above.
    """

    if not isinstance(klass, type):
        raise TypeError('{0} is not a class'.format(klass))
    if isinstance(klass, ABCMeta):
        raise TypeError("{0} can't be a metaclass".format(klass))
    if not issubclass(klass, Iterable):
        raise TypeError('{0} is not iterable'.format(klass))

    extend_type(klass, asq_methods)


def patch_module(module):
    """
    Extends all iterables in the module
    with ``asq.queryables.Queryable`` methods.

    Applies the same as ``patch_type`` to every appropriate member of
    the module.

    Args:
        module: Module object or its __dict__.

    Returns:
        A list of patched classes.
    """

    module_dict = module if isinstance(module, dict) else vars(module)
    patched_types = []
    for klass in module_dict.values():
        if (not isinstance(klass, type) or
                isinstance(klass, ABCMeta) or
                not issubclass(klass, Iterable)):
            continue

        extend_type(klass, asq_methods)
        patched_types.append(klass)
    return patched_types


__all__ = 'patch_type', 'patch_module'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Extend built-in Python collections with LINQ-for-objects style methods

During import, extends all iterables in the built-ins namespace,
``collections`` and ``itertools`` modules with ``asq.queryables.Queryable``
methods.
"""

import collections
import itertools

from .asq_binding import patch_module, patch_type


__author__ = 'Alexander Borzunov'
__version__ = '0.1'


patch_module(__builtins__)
patch_type(type({}.keys()))
patch_type(type({}.values()))
patch_type(type({}.items()))
patch_module(collections)
patch_module(itertools)


__all__ = 'patch_type', 'patch_module'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re
from setuptools import setup


project_dir = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(project_dir, 'dontasq', '__init__.py')) as f:
    version = re.search(r"__version__ = '(.+)'", f.read()).groups()[0]

with open(os.path.join(project_dir, 'README.rst')) as f:
    long_description = f.read()

setup(
    name="dontasq",
    version=version,
    packages=['dontasq'],

    install_requires=['asq>=1.0'],

    author="Alexander Borzunov",
    author_email="borzunov.alexander@gmail.com",

    description='Extend built-in Python collections '
                'with LINQ-for-objects style methods',
    long_description=long_description,
    url="https://github.com/borzunov/dontasq",

    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    license="MIT",
    keywords=['LINQ'],

    test_suite='tests',
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import gc
from collections import MutableMapping


def extend_dict(type_dict, attrs):
    for name, attr in attrs.items():
        if name not in type_dict:
            # TODO: Write in the docs about the IF
            type_dict[name] = attr


def extend_builtin_type(klass, attrs):
    # An interesting hack from:
    #     https://stackoverflow.com/questions/25440694/whats-the-purpose-of-dictproxy

    referents = gc.get_referents(klass.__dict__)
    if not (len(referents) == 1 and isinstance(referents[0], dict)):
        raise ValueError("Can't find writable __dict__ instance for {0}"
                         .format(klass))
    extend_dict(referents[0], attrs)


def extend_type(klass, attrs):
    klass_dict = vars(klass)
    if isinstance(klass_dict, MutableMapping):
        extend_dict(klass_dict, attrs)
    else:
        extend_builtin_type(klass, attrs)


__all__ = 'extend_type'
